/********************************************************************************/
/*         (c) Copyright 2018, Real-Time Innovations, All rights reserved.      */
/*                                                                              */
/* RTI grants Licensee a license to use, modify, compile, and create            */ 
/* derivative works of the software solely for use with RTI Connext DDS.        */ 
/* Licensee may redistribute copies of the software provided that all such      */
/* copies are subject to this license. The software is provided "as is", with   */
/* no warranty of any type, including any warranty for fitness for any purpose. */
/* RTI is under no obligation to maintain or support the software.  RTI shall   */
/* not be liable for any incidental or consequential damages arising out of the */
/* use or inability to use the software.                                        */
/********************************************************************************/
#include <dds/dds.hpp>
#include "GenericSubscriber.h"
#include <rti/core/ListenerBinder.hpp>
#include <rti/topic/rtitopic.hpp>
#include <dds/topic/BuiltinTopic.hpp>

using namespace dds::core;
using namespace dds::core::status;
using namespace dds::domain;
using namespace dds::topic;
using namespace dds::sub;

class GenericSubscriber::GenericSubscriberImpl: public NoOpDataReaderListener<xtypes::DynamicData> 
{
    private:
        using GenericTopic = Topic<xtypes::DynamicData>;
        using GenericReader = DataReader<xtypes::DynamicData>;
        using TopicName = std::string;
        using TopicKind = std::string;
        using TopicType = optional<xtypes::DynamicType>;
        using TopicAndType = std::tuple<TopicName, TopicType>;
        using TopicAndTypeList = std::vector<TopicAndType>;
        using TopicColumnName = std::string;
        using TopicColumnType = xtypes::DynamicType;
        using TopicColumnKey = bool;
        using TopicColumnOptional = bool;
        using TopicColumn = std::tuple<TopicColumnName, TopicColumnType, TopicColumnKey, TopicColumnOptional>;
        using TopicColumnList = std::vector<TopicColumn>;
        using DynamicTopic = xtypes::DynamicData;
        using DynamicTopicList = std::map<TopicName, DynamicTopic>;

        // class used to gather topics from default publisher
        class MyReaderListener: public NoOpDataReaderListener<PublicationBuiltinTopicData> {
            public:
                MyReaderListener(TopicAndTypeList& topics):m_topics(topics)
                {                    
                }

                // incoming data on the builtin topic reader
                void on_data_available(DataReader<PublicationBuiltinTopicData>& rdr)
                {
                    LoanedSamples<PublicationBuiltinTopicData> samples = rdr.take();
                    for (auto& sample: samples) {
                        if (sample.info().valid()) {
                            m_topics.emplace_back(sample.data().topic_name(), sample.data()->type());
                        }
                    }
                }

            private:
                TopicAndTypeList& m_topics;
        };

        // decode the topic kind to a human readable string
        TopicKind topic_kind(const TopicType& topicType) const
        {
            if(xtypes::is_aggregation_type(topicType.get())) {
                return "Aggregation";
            }
            if(xtypes::is_constructed_type(topicType.get())) {
                return "Constructed";
            }
            if(xtypes::is_collection_type(topicType.get())) {
                return "Collection";
            }
            if(xtypes::is_primitive_type(topicType.get())) {
                return "Primitive";
            }

            return "Unknown";
        }

        // wait until the list of topics has been populated
        void wait_for_discovery(void) const
        {
            while(m_discovered_topics.size() == 0) {
                rti::util::sleep(Duration::from_secs(poll_seconds));
            }
        }

        template<typename Funct>
        void process_members(const xtypes::StructType& record, Funct func) const
        {
            for(auto& member: record.members()) {
                func(member);
            }
        }

        // get all of the columns (i.e. members) of the topic type
        void get_topic_columns(const TopicType& topic_type)
        {
            if(m_columns.size() > 0) {
                return;
            }

            if(topic_kind(topic_type) == "Aggregation") {
                auto& record = static_cast<const xtypes::StructType&>(topic_type.get());

                if(xtypes::is_extensible(record) && record.has_parent()) {
                    // get the base members since this is an extensible type
                    process_members(record.parent(), [this](const xtypes::Member& member) {
                        m_columns.emplace_back(member.name(), member.type(), 
                                               member.is_key(), member.is_optional());        
                    } );
                }
                process_members(record, [this](const xtypes::Member& member) {
                    m_columns.emplace_back(member.name(), member.type(), 
                                            member.is_key(), member.is_optional());        
                } );
            }
        }

        // create a dynamic type matching the discovered topic type (for our topic)
        xtypes::StructType create_type_from_topic(const TopicType& topic_type)
        {
            get_topic_columns(topic_type);

            auto type_name = topic_type.get().name();
            xtypes::StructType whatever(type_name);
            for(const auto& column: m_columns) {
                whatever.add_member(xtypes::Member(std::get<0>(column), 
                                                   std::get<1>(column))
                                                    .key(std::get<2>(column))
                                                    .optional(std::get<3>(column))
                                                  );
            }
 
            return whatever;
        }

        // represent given dynamic data value as a string
        std::string value_as_string(const xtypes::DynamicData& data, const std::string& member) const
        {
            std::stringstream buffer;

            switch(data.member_info(member).member_kind().underlying()) {
                case xtypes::TypeKind_def::BOOLEAN_TYPE:
                    buffer << data.value<bool>(member);
                    return buffer.str();                 
                case xtypes::TypeKind_def::UINT_8_TYPE: 
                    buffer << data.value<uint8_t>(member);
                    return buffer.str();                 
                case xtypes::TypeKind_def::INT_16_TYPE: 
                    buffer << data.value<int16_t>(member);
                    return buffer.str();                 
                case xtypes::TypeKind_def::UINT_16_TYPE: 
                    buffer << data.value<uint16_t>(member);
                    return buffer.str();                 
                case xtypes::TypeKind_def::INT_32_TYPE: 
                    buffer << data.value<int32_t>(member);
                    return buffer.str();                 
                case xtypes::TypeKind_def::UINT_32_TYPE: 
                    buffer << data.value<uint32_t>(member);
                    return buffer.str();                 
                case xtypes::TypeKind_def::INT_64_TYPE: 
                    buffer << data.value<int64_t>(member);
                    return buffer.str();                 
                case xtypes::TypeKind_def::UINT_64_TYPE: 
                    buffer << data.value<uint64_t>(member);
                    return buffer.str();                 
                case xtypes::TypeKind_def::FLOAT_32_TYPE: 
                    buffer << data.value<float>(member);
                    return buffer.str();                 
                case xtypes::TypeKind_def::FLOAT_64_TYPE: 
                    buffer << data.value<double>(member);
                    return buffer.str();                 
                case xtypes::TypeKind_def::FLOAT_128_TYPE: 
                    buffer << "Unsupported Floating point type";
                    return buffer.str();                 
                case xtypes::TypeKind_def::CHAR_8_TYPE: 
                    buffer << data.value<char>(member);
                    return buffer.str();                 
                case xtypes::TypeKind_def::ENUMERATION_TYPE:
                    buffer << data.value<uint32_t>(member);
                    return buffer.str(); 
                case xtypes::TypeKind_def::STRING_TYPE: 
                    buffer << data.value<std::string>(member);
                    return buffer.str();                 
                case xtypes::TypeKind_def::WSTRING_TYPE: 
                    return std::string();                 
                case xtypes::TypeKind_def::UNION_TYPE: 
                    return std::string();                 
                case xtypes::TypeKind_def::STRUCTURE_TYPE:  
                    buffer << data.value<bool>(member);
                    return buffer.str();  
                case xtypes::TypeKind_def::ARRAY_TYPE:
                    return std::string();
                case xtypes::TypeKind_def::SEQUENCE_TYPE: {
                    std::vector<uint8_t> values;

                    data.get_values(member, values);
                    int i=0;
                    for(auto& value: values) {
                        if(++i > 16) {
                            break;
                        }
                        buffer << std::hex << static_cast<int>(value);
                    }
                    return buffer.str();
                }
                default:
                    return std::string();
            }
        }

        // incoming data from the topic we subscribed to
        void on_data_available(DataReader<xtypes::DynamicData>& rdr)
        {
            LoanedSamples<xtypes::DynamicData> samples = rdr.take();
            for (auto& sample: samples)  {
                if (sample.info().valid()) {
                    for(auto& column: m_columns) {
                        std::cout << "  " << std::setw(16) << value_as_string(sample.data(), 
                                                                              std::get<0>(column));
                    }
                    std::cout << std::endl;
                }
            }
        }

        // print out header based on discovered topic type members
        void header(TopicType topicType)
        {
            get_topic_columns(topicType);

            std::cout << " ";
            for(auto& column: m_columns) {
                std::cout << std::setw(12) << std::get<0>(column) << "     |";
            }
            std::cout << std::endl; 
        }
 
    public:
        // construct a subscriber
        GenericSubscriberImpl(int dom_id, bool verbose): m_participant(dom_id),
                                                m_ear(m_discovered_topics),
                                                m_builtin_subscriber(builtin_subscriber(m_participant)),
                                                m_verbose(verbose)
        {
            // locate the built-in subscriber for the topic names
            find<DataReader<PublicationBuiltinTopicData>>(m_builtin_subscriber, 
                                                          dds::topic::publication_topic_name(), 
                                                          std::back_inserter(m_rdr_list));
            // attach listener to all readers
            for(auto& rdr: m_rdr_list) {
                rdr.listener(&m_ear, StatusMask::data_available());
            }

            // only now that we have our listeners attached are we ready to enable the participant
            m_participant.enable();
        }

        // list all available topics we discovered on the domain
        void list_topics(void)
        {
            wait_for_discovery();
            
            std::cout << "Number of topics discovered: " << m_discovered_topics.size() << std::endl;
            for(auto topic: m_discovered_topics) {
                std::cout << "Discovered topic: " << std::get<0>(topic) << std::endl;
            }
        }

        void receive(const TopicName& topic_name)
        {
            TopicType topic_type;

            wait_for_discovery();

            for(auto topic: m_discovered_topics) {
                if(std::get<0>(topic) == topic_name) {
                    topic_type = std::get<1>(topic);
                    if(topic_type.is_set()) {
                        std::cout << "Enabling subscribing on topic " << topic_name
                                  << " which is of type " << topic_type.get().name() 
                                  << " which is a kind of " << topic_kind(topic_type) << std::endl;

                        m_topic = GenericTopic(m_participant, topic_name, 
                                               create_type_from_topic(topic_type));
                        m_reader = GenericReader(Subscriber(m_participant), m_topic);
                        break;  // only support subscribing to a single topic
                    }
                }
            }

            if(m_reader.is_nil()) {
                return;
            }

            auto scoped_listener = rti::core::bind_listener(m_reader, this, 
                                                            dds::core::status::StatusMask::data_available());

            // display headers 
            header(topic_type);

            // while there are writers available collect and display data
            while(true) {
                rti::util::sleep(Duration::from_secs(poll_seconds));
            }
        }

    private:
        TopicAndTypeList                                        m_discovered_topics;

        DomainParticipant                                       m_participant;
        MyReaderListener                                        m_ear;
        Subscriber                                              m_builtin_subscriber;
        std::vector<DataReader<PublicationBuiltinTopicData>>    m_rdr_list;    
        GenericTopic                                            m_topic = nullptr;
        GenericReader                                           m_reader = nullptr;

        bool                                                    m_verbose;
        TopicColumnList                                         m_columns;

        static const int                                        poll_seconds = 1;
};

GenericSubscriber::GenericSubscriber(int domain_id, bool verbose)
    : impl(std::make_unique<GenericSubscriberImpl>(domain_id, verbose))
{
}

GenericSubscriber::~GenericSubscriber()
{
}

void GenericSubscriber::list_topics(void)
{
    impl->list_topics();
}

void GenericSubscriber::receive(const std::string& topic_name)
{
    impl->receive(topic_name);
}